#!/bin/bash
#===================================================================================================================================
#Author: Badrinath Nagarajan
#------
#Date: Jan 2011
#-----
#Purpose: -Download CMC high resolution real-time operational analysis
#         -convert the grib2 file to RPN FST format for GEM real-time fcst initialization
#
#Requisites: 1) CMC GRIB package to convert grib1 to FST format 
#               Download: http://collaboration.cmc.ec.gc.ca/science/rpn.comm/cgi-bin/download.php?request=cmcgrib_Linux.tar.Z
#
#            2) cnvgrib package from NCEP
#               Download: http://www.nco.ncep.noaa.gov/pmb/codes/GRIB2/g2lib-1.2.1.tar
#                         http://www.nco.ncep.noaa.gov/pmb/codes/GRIB2/w3lib-1.9.tar
#                         http://www.nco.ncep.noaa.gov/pmb/codes/GRIB2/cnvgrib-1.2.1.tar
#                         http://www.nco.ncep.noaa.gov/pmb/codes/GRIB2/Build_cnvgrib.pdf (Documentation) 
# 
#               OR
#               cnvgrib 32 and 64 bit binaries (stand alone) are provided in PRESENT DIRECTORY/../tools/grib 
#               May be you want to use these.
#
#Environment Variables: 1) CMCGRIB --> Location of CMC grib install that contains grib_decode and dictionary.grib.
#                                      (e.g., /d1/badri/model/ICLBC/cmcgrib_cvt-3.4.1)
#                       2) CMCDICT --> Location of customized dictionary files to be used to convert grib1 to FST
#                                      (i.e., dictionary.grib.only_i7_1197, dictionary.grib.only_i7_1199_1198, dictionary.grib.only_i9
#                                             and dictionary.grib.all_except_i7_i9 )
#
#Usage:
#       fetch_cmcanalysis_convert_grib2fst.sh --analysis_hour=00 
# 
#       fetch_cmcanalysis_convert_grib2fst.sh --analysis_hour=12
#
#       for 00 and 12 UTC initialized forecasts respectively
#
#Notes:
#I) I7 and I9  (or Sea-ice and Glacier) temperature fields.
#----------------------------------------------------------------------
#I9 	ip3=1 	Superficial temperature of glaciers                   K
#       Alternatives [If field not present]:
#       I9 (ip1=0,ip3=1) or
#       I0 (ip1=convip(1)) or
#       TS (ip1=0)
#I9 	ip3=2 	Deep temperature of glaciers                          K
#       Alternatives [If field not present]:
#       I9(ip1=0,ip3=2) or
#       TP(ip1=0)
#I7 	ip3=1 	Sea-Ice temperature                                   K
#       Alternatives [If field not present]:
#       I7(ip1=0) or
#       I0(ip1=convip(1)) or
#       TS(ip1=0)
#I7 	ip3=2 	Deap Ground temperature, layer 1                      K
#       Alternatives [If field not present]:
#       TP(ip1=0) 
#I7 	ip3=3 	Deap Ground temperature, layer 2                      K
#       Alternatives [If field not present]:
#
#II)- We obtain I7 and I9 from I0 given out by CMC as part of their analysis (00 and 12 UTC)
#   - TSOIL_SFC and TSOIL_DBLL are I0 (1199 and 1198) respectively 
#===================================================================================================================================

#Check for CMCGRIB and CMCDICT
if [ ${#CMCGRIB} == 0 ] || [ ${#CMCDICT} == 0 ]; then
 echo "====================================================================="
 echo "Environment Variable CMCGRIB AND/OR CMCDICT not set by user"
 echo "export CMCGRIB="location where your CMCGRIB install files are located"
 echo "export CMCDICT="location where your:"
 echo "dictionary.grib.only_i9_1197, dictionary.grib.only_i9_1199_1198, dictionary.grib.only_i7, dictionary.grib.all_except_i7_i9 are located"
 echo "Exiting .."
 echo "====================================================================="
 exit
fi

#User options
while [ $1 ] ; do
   case $1 in
      --analysis_hour=*)
         analysis_hour=`echo $1 | cut -d= -f2`
      ;;
   esac
   shift
done

#Default values
: ${analysis_hour:=00}
#User options

#Initiliaze
HH=${analysis_hour}

#Surface & deep ground temperature field prefixes
PREFIX_SFC="TSOIL_SFC"
PREFIX_DEEP_SOIL="TSOIL_DBLL"
export PREFIX_SFC PREFIX_DEEP_SOIL 

#
#Download grib2 data from CMC 
#============================
#Download HH UTC analysis fields ONLY (controlled by 12_P000 and -l 1)
wget --no-directories -nH --accept ${HH}_P000.grib2 -o log -l 1 --recursive "http://dd.weatheroffice.ec.gc.ca/model_gem_global/high_resolution/grib2/lat_lon/${HH}/000/"

#NOTE: 
#--no-directories: put stuff in current dir, 
#-nH: do not create host dir structure either
#--accept grib2: download only grib2 files
#-o log: wget writes its progress to a log file

#Convert grib2 to grib1 (00UTC)
#======================
#Exclude Several fields 
ls -1 *${HH}_P000.grib2|awk 'BEGIN{FS=".grib2"}{if (index($1,"_TGL") == 0      &&  \
                                                    index($1,"_WIND") == 0     &&  \
                                                    index($1,"_WDIR") == 0     &&  \
                                                    index($1,"_VVEL") == 0     &&  \
                                                    index($1,"_DEPR") == 0     &&  \
                                                    index($1,"_ABSV") == 0     &&  \
                                                    index($1,"_HGT_SFC")==0    &&  \
                                                    index($1,"_TCDC_SFC")==0   &&  \
                                                    index($1,"_SPFH_SFC")==0   &&  \
                                                    index($1,"_ULWRF_NTAT")==0)    \
                                                    print "cnvgrib -g21 "$1".grib2 "$1".grib1"}' > doit.sh
. ./doit.sh

#grib1 to fst conversion
#=======================
#This process is repeated: a) FOR ALL FIELDS EXCEPT I7,I9, b) I7 - 2 levels c) I9 - 3 levels.
#--------------------------------------------------------------------------------------------
#a)  A cranky way to handle ip2 generation (month info) for AL and TM fields
MM=$(ls -1 *${HH}_P000.grib2 |head -n 1 |awk 'BEGIN{FS=".grib2"}{print $1}'|awk 'BEGIN{FS="_"}{for( i = 1 ; i <= NF ; i++ ) if (length($i) == 10) print substr($i,5,2) }')
export MM
echo "cat << EOF"                                   > a
cat dictionary.grib.all_except_i7_i9.tmp           >> a
echo "EOF"                                         >> a
chmod u+x a
./a >  dictionary.grib.all_except_i7_i9
ls -1 *${HH}_P000.grib1 |awk '{print "${CMCGRIB}/grib_decode  -dict ${CMCDICT}/dictionary.grib.all_except_i7_i9  -v -o out.fst -i "$1}' > doit1.sh
. ./doit1.sh

#b) 
ls -1       *${PREFIX_SFC}*${HH}_P000.grib1 |awk '{print "${CMCGRIB}/grib_decode  -dict ${CMCDICT}/dictionary.grib.only_i9  -v -o out.fst -i "$1}' > doit1.sh
. ./doit1.sh
ls -1 *${PREFIX_DEEP_SOIL}*${HH}_P000.grib1 |awk '{print "${CMCGRIB}/grib_decode  -dict ${CMCDICT}/dictionary.grib.only_i9  -v -o out.fst -i "$1}' > doit1.sh
. ./doit1.sh

#c)
ls -1       *${PREFIX_SFC}*${HH}_P000.grib1 |awk '{print "${CMCGRIB}/grib_decode  -dict ${CMCDICT}/dictionary.grib.only_i7_1199_1198  -v -o out.fst -i "$1}' > doit1.sh
. ./doit1.sh
ls -1 *${PREFIX_DEEP_SOIL}*${HH}_P000.grib1 |awk '{print "${CMCGRIB}/grib_decode  -dict ${CMCDICT}/dictionary.grib.only_i7_1199_1198  -v -o out.fst -i "$1}' > doit1.sh
. ./doit1.sh
ls -1 *${PREFIX_DEEP_SOIL}*${HH}_P000.grib1 |awk '{print "${CMCGRIB}/grib_decode  -dict ${CMCDICT}/dictionary.grib.only_i7_1197       -v -o out.fst -i "$1}' > doit1.sh
. ./doit1.sh


YYYYMMDD=$(ls -1 *${HH}_P000.grib1 |head -n 1 |awk 'BEGIN{FS=".grib1"}{print $1}'|awk 'BEGIN{FS="_"}{for( i = 1 ; i <= NF ; i++ ) if (length($i) == 10) print substr($i,1,8) }')
HH=$(ls -1 *${HH}_P000.grib1 |head -n 1 |awk 'BEGIN{FS=".grib1"}{print $1}'|awk 'BEGIN{FS="_"}{for( i = 1 ; i <= NF ; i++ ) if (length($i) == 10) print substr($i,9,2)}')
mv out.fst ${YYYYMMDD}"."${HH}"0000"

#Cleanup
#=======
rm -f ./doit.sh ./doit1.sh *.grib1 ./a  dictionary.grib.all_except_i7_i9
mkdir ${YYYYMMDD}"."${HH}"0000".grib2
#Put away *.grib2 files we downloaded. User can delete it or archive it 
mv log CMC*.grib2 ${YYYYMMDD}"."${HH}"0000".grib2
