#!/bin/bash
# Purpose: Script to launch gemntr (pre-processor) for GEMDM model
# Author: Badrinath Nagarajan & Ruben Santos
# Date:   Dec 2010-Jan 2011
# Disclaimer: Provided "as is" and no warranty whatsoever
#
# Pre-requisites: 
# Is mpiexec in your PATH ? Otherwise set it
#
# Usage: 
#       - Interactively from the headnode of your cluster OR
#       - qsub this script
#
#--Environment Variables for GEMNTR and GEMDM
#============================================
. ./envvars.sh

#-- GEMDM needs TMPDIR (so set it up)
mkdir $EXECDIR/tmp > /dev/null 2> /dev/null
export TMPDIR=${EXECDIR}/tmp

#-- Get number of CPUs/processors from npex,npey of gem_settings file
#=========================================================================================
cd $EXECDIR/process
ncpu=`grep npex model_settings |awk 'BEGIN{FS=","}{print $1" "$2}'|awk '{print expr $3*$6}'`

#--MPI setup (MPICH 2 or OPENMPI)
#================================
#--Figure out if using mpich2 or openmpi
path_mpiexec=$(which mpiexec)
#We delete trailing/leading quotes using tr -d 
lastword=$(file ${path_mpiexec}|awk 'BEGIN{FS=":"}{print $2}'|awk '{print $4}'|tr -d "\' \`")

if [ ${lastword} == "orterun" ]; then
 openmpi=1
 mpich2=0
 echo "======================================="
 echo "Found we are using mpiexec from openmpi"
 echo "======================================="
elif [ ${lastword} == "mpiexec.py" ]; then
 openmpi=0
 mpich2=1
 echo "====================================="
 echo "Found we are using mpiexec from mpich"
 echo "====================================="
fi

cd $EXECDIR

#-- If mpich2 cleanup first
if [ ${mpich2} == 1 ]; then
 mpdallexit
 mpdcleanup
fi

#-- Prepare the machine file using PBS_NODEFILE variable
#-- or the user should give us the machine file.
#=========================================================
if [ ${#PBS_NODEFILE} -gt 0 ]; then
 cat $PBS_NODEFILE > hosts.list
 echo "====================================== "
 echo "Generated hosts.list using PBS_NODEFILE"
 echo "======================================"
elif [ -f ./machines.txt ]; then
 ln -s machines.txt hosts.list > /dev/null 2> /dev/null
 echo "============================================"
 echo "Generated hosts.list using user machines.txt"
 echo "============================================"
else
 echo 
 echo "=============================================================="
 echo "Please provide machines.txt containing node names one per line"
 echo "These are the nodes on which the model will run .."
 echo "Exiting .."
 echo "=============================================================="
 echo
 exit
fi

#--Generate no of hosts (note some hosts have 2 or more cpus!)
#--Sort and then get the unique hosts
#=====================================================
sort hosts.list > sorted;uniq sorted > uniq
nhosts=`wc -l uniq|awk '{print $1}'`
rm -f ./uniq ./sorted

#--Cleanup hanging gemdm.Abs (left overs)
#========================================
#NOTE:
#     -It may be that your cluster/desktop is not setup to ssh into without
#      asking for a password. In that case do the following:
#      step-1: ssh-keygen
#      step-2: cp rsa.pub  authorized_keys
#NOTE:
#-----------------------------------------
cat hosts.list |awk '{print "ssh "$1" killall -e gemdm.Abs"}' > dogemdmabs.sh
chmod u+x do*.sh
./dogemdmabs.sh > /dev/null 2> /dev/null
rm -f dogemdmabs.sh  > /dev/null

#--Cleanup process/00-00 etc directories
#=======================================
rm -f ${EXECDIR}/process/0*/* > /dev/null

#--Launch mpdboot for mpich2
#===========================
if [ ${mpich2} == 1 ]; then
 mpdboot --totalnum=$nhosts --file=./hosts.list --rsh="ssh" --user=$USER --ncpus=$ncpu
 mpdtrace -l
fi

#--Launch MPI GEMDM
#==================
#Some old dual processor machines needed the ":1". 
#cat hosts.list |awk '{print $1":1"}' > newhosts.list
#mv newhosts.list hosts.list

mpiexec -machinefile ./hosts.list -n $ncpu ./gemdm.Abs > gemdm.list 2> gemdm.2

#--Useful for debugging (mpich2 only) MPI debug
#==============================================
#mpiexec -machinefile ./hosts.list -gdb -n $ncpu ./gemdm.Abs

#--Shutdown mpdboot
#==================
if [ ${mpich2} == 1 ]; then
 mpdallexit
 mpdcleanup
fi
